{
  "type": "record",
  "Aspect": {
    "name": "mlMetric"
  },
  "name": "MLMetric",
  "namespace": "com.linkedin.pegasus2avro.ml.metadata",
  "fields": [
    {
      "type": "string",
      "name": "name",
      "doc": "Name of the mlMetric"
    },
    {
      "type": [
        "null",
        "string"
      ],
      "name": "description",
      "default": null,
      "doc": "Documentation of the mlMetric"
    },
    {
      "type": [
        "null",
        "string"
      ],
      "name": "value",
      "default": null,
      "doc": "The value of the mlMetric"
    },
    {
      "type": [
        "null",
        "long"
      ],
      "name": "createdAt",
      "default": null,
      "doc": "Date when the mlMetric was developed"
    }
  ],
  "doc": "Properties associated with an ML Metric"
}